@@ -52,6 +52,7 @@ def marg_gates_est(forest, fill_y_sample, pred_sample, v_dict, c_dict,
     any_plots_amgate = False
     c_dict_mgate = copy.deepcopy(c_dict)
     c_dict_mgate['with_output'] = False       # reduce unnecessary infos
+    c_dict_mgate['iate_se_flag'] = True
     if v_dict['z_name_mgate'] and c_dict['with_output']:
         any_plots_mgate = mgate_function(
             forest, fill_y_sample, pred_sample, v_dict, c_dict, x_name_mcf,
@@ -93,13 +94,18 @@ def amgate_function(forest, fill_y_sample, pred_sample, v_dict, c_dict,
     x_name_mcf : Names from MCF procedure.
     var_x_type : List of int. Type of feature.
     var_x_values : List of List of float or int. Values of features.
-    c_dict_mgate: Dict. Differs only for 'with_output' from c_dict.
+    c_dict_mgate: Dict. Differs only for 'with_output' (t) from c_dict.
 
     Returns
     -------
     any_plots_done : Bool.
 
     """
+    if c_dict['gatet_flag']:
+        c_dict_mgate['gatet_flag'] = False
+        c_dict_mgate['atet_flag'] = False
+        if c_dict['with_output']:
+            print('No treatment specific effects for MGATE and AMGATE.')
     any_plots_done = False
     if c_dict['with_output']:
         print()
@@ -127,8 +133,9 @@ def amgate_function(forest, fill_y_sample, pred_sample, v_dict, c_dict,
             w_ate_iate, _, _, ate_z, ate_se_z, _ = mcf_ate.ate_est(
                     weights, new_predict_file, y_f, cl_f, w_f, v_dict,
                     c_dict_mgate)
+            c_dict_mgate['with_output'] = c_dict['with_output']
             gate_est(weights, new_predict_file, y_f, cl_f, w_f, v_dict,
-                     c_dict, var_x_type, var_x_values, w_ate_iate, ate_z,
+                     c_dict_mgate, var_x_type, var_x_values, w_ate_iate, ate_z,
                      ate_se_z, amgate_flag=True)
             os.remove(new_predict_file)  # Delete new file
     v_dict['z_name'] = z_name_old
@@ -164,6 +171,8 @@ def mgate_function(
             print('Marginal variable predictive plots')
         else:
             print('Marginale GATEs evaluated at median (MGATES)', '\n')
+        if c_dict_mgate['choice_based_yes']:
+            print('Choice based sampling deactivated for MGATES.')
     ref_values, eva_values = ref_vals_margplot(
         pred_sample, var_x_type, var_x_values,
         with_output=c_dict['with_output'], ref_values_needed=True,
@@ -175,6 +184,8 @@ def mgate_function(
     c_dict_mgate['w_yes'] = False   # Weighting not needed here
     with_output_old = c_dict_mgate['with_output']
     c_dict_mgate['with_output'] = False
+    choice_based_yes_old = c_dict_mgate['choice_based_yes']
+    c_dict_mgate['choice_based_yes'] = False
     for vname in v_dict['z_name_mgate']:
         if vname in x_name_mcf:
             if c_dict['with_output'] and c_dict['verbose']:
@@ -225,6 +236,7 @@ def mgate_function(
                   '(equally weighted).')
     c_dict_mgate['w_yes'] = w_yes_old
     c_dict_mgate['with_output'] = with_output_old
+    c_dict_mgate['choice_based_yes'] = choice_based_yes_old
     return any_plots_done
 
 
@@ -262,15 +274,9 @@ def plot_marginal(pred, pred_se, names_pred, x_name, x_values_in, x_type,
                 titel = 'MGATEMATE ' + x_name + names_pred[idx][:-9]
             else:
                 titel = 'MGATE ' + x_name + names_pred[idx][:-5]
-        # pred_temp = pred[:, idx].flatten()
-        # pred_se_temp = pred_se[:, idx].flatten()
         pred_temp = pred[:, idx]
         pred_se_temp = pred_se[:, idx]
         if x_type == 0 and len(x_values) > c_dict['no_filled_plot']:
-            # pred_temp = gp.moving_avg_mean_var(pred_temp, x_values, 3,
-            #                                    False)[0]
-            # pred_se_temp = gp.moving_avg_mean_var(pred_se_temp, x_values, 3,
-            #                                       False)[0]
             pred_temp = gp_est.moving_avg_mean_var(pred_temp, 3, False)[0]
             pred_se_temp = gp_est.moving_avg_mean_var(
                 pred_se_temp, 3, False)[0]
@@ -281,14 +287,19 @@ def plot_marginal(pred, pred_se, names_pred, x_name, x_values_in, x_type,
         upper = pred_temp + pred_se_temp * conf_int
         lower = pred_temp - pred_se_temp * conf_int
         fig, axs = plt.subplots()
-        # label_t = names_pred[idx]
         label_u = 'Upper ' + str(round(c_dict['fig_ci_level'] * 100)) + '%-CI'
         label_l = 'Lower ' + str(round(c_dict['fig_ci_level'] * 100)) + '%-CI'
         label_ci = str(c_dict['fig_ci_level'] * 100) + '%-CI'
         if regrf:
             label_m = 'Conditional prediction'
         else:
-            label_m = 'MGATE'
+            if minus_ate:
+                label_m = 'MGATE-ATE'
+                label_0 = '_nolegend_'
+                line_0 = '_-k'
+                zeros = np.zeros_like(pred_temp)
+            else:
+                label_m = 'MGATE'
         if x_type == 0 and len(x_values) > c_dict['no_filled_plot']:
             axs.plot(x_values, pred_temp, label=label_m, color='b')
             axs.fill_between(x_values, upper, lower, alpha=0.3, color='b',
@@ -305,11 +316,14 @@ def plot_marginal(pred, pred_se, names_pred, x_name, x_values_in, x_type,
                 connect_x[0] = i_lab
                 connect_x[1] = i_lab
                 axs.plot(connect_x, connect_y, 'b-', linewidth=0.7)
-            axs.plot(x_values, pred_temp, middle + 'b', label=label_m)
+            axs.plot(x_values, pred_temp, middle +'b', label=label_m)
             axs.plot(x_values, upper, u_line + 'b', label=label_u)
             axs.plot(x_values, lower, l_line + 'b', label=label_l)
+            if minus_ate:
+                axs.plot(x_values, zeros, line_0, label=label_0)
         if c_dict['with_output']:
             print_mgate(pred_temp, pred_se_temp, titel, x_values)
+            gp_est.print_se_info(c_dict['cluster_std'], c_dict['se_boot_gate'])
         axs.set_ylabel(label_m)
         axs.legend(loc='lower right', shadow=True,
                    fontsize=c_dict['fig_fontsize'])
@@ -326,8 +340,10 @@ def plot_marginal(pred, pred_se, names_pred, x_name, x_values_in, x_type,
         upper = upper.reshape(-1, 1)
         lower = lower.reshape(-1, 1)
         pred_temp = pred_temp.reshape(-1, 1)
-        effects_et_al = np.concatenate((upper, pred_temp, lower), axis=1)
-        cols = ['upper', 'effects', 'lower']
+        x_values_np = np.array(x_values, copy=True).reshape(-1, 1)
+        effects_et_al = np.concatenate((upper, pred_temp, lower, x_values_np),
+                                        axis=1)
+        cols = ['upper', 'effects', 'lower', 'x_values']
         datasave = pd.DataFrame(data=effects_et_al, columns=cols)
         gp.delete_file_if_exists(file_name_csv)
         datasave.to_csv(file_name_csv, index=False)
@@ -641,18 +657,18 @@ def gate_est(weights_all, pred_data, y_dat, cl_dat, w_dat, v_dict, c_dict,
             maxworkers = c_dict['no_parallel']
         if weights_all2 is None:
             maxworkers = round(maxworkers / 2)
-        # if maxworkers > no_of_zval:
-        #     maxworkers = no_of_zval
         if not maxworkers > 0:
             maxworkers = 1
     if c_dict['with_output'] and c_dict['verbose']:
         print('Number of parallel processes: ', maxworkers, flush=True)
     if c_dict['mp_with_ray']:
         if c_dict['mem_object_store_3'] is None:
-            ray.init(num_cpus=maxworkers, include_dashboard=False)
+            if not ray.is_initialized():
+                ray.init(num_cpus=maxworkers, include_dashboard=False)
         else:
-            ray.init(num_cpus=maxworkers, include_dashboard=False,
-                     object_store_memory=c_dict['mem_object_store_3'])
+            if not ray.is_initialized():
+                ray.init(num_cpus=maxworkers, include_dashboard=False,
+                         object_store_memory=c_dict['mem_object_store_3'])
             if c_dict['with_output'] and c_dict['verbose']:
                 print("Size of Ray Object Store: ",
                       round(c_dict['mem_object_store_3']/(1024*1024)), " MB")
@@ -688,22 +704,6 @@ def gate_est(weights_all, pred_data, y_dat, cl_dat, w_dat, v_dict, c_dict,
         pot_y_var = np.empty(dim_all)
         pot_y_mate = np.empty(dim_all)
         pot_y_mate_var = np.empty(dim_all)
-        # if c_dict['no_parallel'] < 1.5:
-        #     maxworkers = 1
-        # else:
-        #     if c_dict['mp_automatic']:
-        #         maxworkers = gp.find_no_of_workers(c_dict['no_parallel'],
-        #                                            c_dict['sys_share']/2)
-        #     else:
-        #         maxworkers = c_dict['no_parallel']
-        #     if weights_all2 is None:
-        #         maxworkers = round(maxworkers / 2)
-        #     if maxworkers > no_of_zval:
-        #         maxworkers = no_of_zval
-        #     if not maxworkers > 0:
-        #         maxworkers = 1
-        # if c_dict['with_output']:
-        #     print('Number of parallel processes: ', maxworkers, flush=True)
         if maxworkers == 1:
             for zj_idx in range(no_of_zval):
                 results_fut_zj = gate_zj(
@@ -721,16 +721,12 @@ def gate_est(weights_all, pred_data, y_dat, cl_dat, w_dat, v_dict, c_dict,
                      w_gate, w_gate_unc, w_censored, results_fut_zj, zj_idx)
         else:
             if c_dict['mp_with_ray']:
-                # ray.init(num_cpus=maxworkers, include_dashboard=False,
-                #          object_store_memory=c_dict['mem_object_store'])
-                # weights_all_ref = ray.put(weights_all)
-                tasks = [ray_gate_zj_mp.remote(
+                still_running = [ray_gate_zj_mp.remote(
                          z_values[zj_idx], zj_idx, y_dat, cl_dat,
                          w_dat, z_p, d_p, w_p, z_name_j, weights_all_ref,
                          w_gate0_dim, w_ate, i_d_val, t_probs, no_of_tgates,
                          no_of_out, c_dict, n_y, bandw_z, kernel, save_w_file,
                          z_smooth) for zj_idx in range(no_of_zval)]
-                still_running = list(tasks)
                 while len(still_running) > 0:
                     finished, still_running = ray.wait(still_running)
                     finished_res = ray.get(finished)
@@ -742,7 +738,6 @@ def gate_est(weights_all, pred_data, y_dat, cl_dat, w_dat, v_dict, c_dict,
                         w_gate, w_gate_unc, w_censored = assign_w(
                             w_gate, w_gate_unc, w_censored, results_fut_idx,
                             results_fut_idx[6])
-                # ray.shutdown()
             else:
                 with futures.ProcessPoolExecutor(max_workers=maxworkers
                                                  ) as fpp:
@@ -841,9 +836,12 @@ def gate_est(weights_all, pred_data, y_dat, cl_dat, w_dat, v_dict, c_dict,
                           ref_pop_lab[a_idx])
                     gp_mcf.print_effect_z(ret_gate, ret_gate_mate, z_values,
                                           gate_str)
+                    gp_est.print_se_info(c_dict['cluster_std'],
+                                         c_dict['se_boot_gate'])
         if c_dict['with_output']:   # figures
             primes = gp.primes_list()
             for a_idx, a_lab in enumerate(ref_pop_lab):
+                gatet_yes = False if a_idx == 0 else True
                 for o_idx, o_lab in enumerate(v_dict['y_name']):
                     for t_idx, t_lab in enumerate(treat_comp_label):
                         for e_idx, e_lab in enumerate(effect_type_label):
@@ -866,13 +864,21 @@ def gate_est(weights_all, pred_data, y_dat, cl_dat, w_dat, v_dict, c_dict,
                             make_gate_figures(
                                 e_lab + z_name + a_lab + o_lab + t_lab, z_name,
                                 z_values_f, z_type_l, effects, ste, c_dict,
-                                ate_f, ate_f_se, amgate_flag, z_smooth)
+                                ate_f, ate_f_se, amgate_flag, z_smooth,
+                                gatet_yes=gatet_yes)
         if c_dict['with_output']:
             print('-' * 80)
         gate[z_name_j] = gate_z
         gate_se[z_name_j] = gate_z_se
     if c_dict['mp_with_ray']:
-        ray.shutdown()
+        if 'refs' in c_dict['_mp_ray_del']:
+            del weights_all_ref
+        # if 'remote' in c_dict['_mp_ray_del']:
+        #     del tasks
+        if 'rest' in c_dict['_mp_ray_del']:
+            del finished_res, finished
+        if c_dict['_mp_ray_shutdown']:
+            ray.shutdown()
     if files_to_delete:  # delete temporary files
         for file in files_to_delete:
             os.remove(file)
@@ -1199,7 +1205,7 @@ def wald_test(z_name, no_of_zval, w_gate, y_dat, w_dat, cl_dat, a_idx, o_idx,
 
 def make_gate_figures(titel, z_name, z_values, z_type, effects, stderr,
                       c_dict, ate=0, ate_se=None, am_gate=False,
-                      z_smooth=False):
+                      z_smooth=False, gatet_yes=False):
     """Generate the figures for GATE results.
 
     Parameters
@@ -1240,12 +1246,18 @@ def make_gate_figures(titel, z_name, z_values, z_type, effects, stderr,
                 if am_gate:
                     label_m = 'AMGATE-ATE'
                 else:
-                    label_m = 'GATE-ATE'
+                    if gatet_yes:
+                        label_m = 'GATET-ATE'
+                    else:    
+                        label_m = 'GATE-ATE'
             else:
                 if am_gate:
                     label_m = 'AMGATE'
                 else:
-                    label_m = 'GATE'
+                    if gatet_yes:
+                        label_m = 'GATET'
+                    else:
+                        label_m = 'GATE'
             axs.plot(z_values, effects, label=label_m, color='b')
             axs.fill_between(z_values, upper, lower, alpha=0.3, color='b',
                              label=label_ci)
@@ -1256,7 +1268,7 @@ def make_gate_figures(titel, z_name, z_values, z_type, effects, stderr,
                     alpha=0.3, color='r', label=label_ci)
                 label_ate = 'ATE'
             else:
-                label_ate = 'Zero'
+                label_ate = '_nolegend_'
             axs.plot(z_values, ate, line_ate, label=label_ate)
             axs.set_ylabel(label_m)
             axs.legend(loc='lower right', shadow=True,
@@ -1292,18 +1304,20 @@ def make_gate_figures(titel, z_name, z_values, z_type, effects, stderr,
             gate_str = 'AMGATE'
         else:
             gate_str = 'GATE'
+        ate_label = 'ATE'
     else:
         if am_gate:
             gate_str = 'AMGATE-ATE'
         else:
             gate_str = 'GATE-ATE'
+        ate_label = '_nolegend_'
     axe.plot(z_values, effects, e_line + 'b', label=gate_str)
     axe.set_ylabel(gate_str)
     label_u = 'Upper ' + str(round(c_dict['fig_ci_level'] * 100)) + '%-CI'
     label_l = 'Lower ' + str(round(c_dict['fig_ci_level'] * 100)) + '%-CI'
     axe.plot(z_values, upper, u_line + 'b', label=label_u)
     axe.plot(z_values, lower, l_line + 'b', label=label_l)
-    axe.plot(z_values, ate, '-' + 'k', label='ATE')
+    axe.plot(z_values, ate, '-' + 'k', label=ate_label)
     if ate_se is not None:
         axe.plot(z_values, ate_upper, '--' + 'k', label=label_u)
         axe.plot(z_values, ate_lower, '--' + 'k', label=label_l)
@@ -1321,15 +1335,19 @@ def make_gate_figures(titel, z_name, z_values, z_type, effects, stderr,
     effects = effects.reshape(-1, 1)
     upper = upper.reshape(-1, 1)
     lower = lower.reshape(-1, 1)
+    z_values_np = np.array(z_values, copy=True).reshape(-1, 1)
     if ate_se is not None:
         ate_upper = ate_upper.reshape(-1, 1)
         ate_lower = ate_lower.reshape(-1, 1)
         effects_et_al = np.concatenate((upper, effects, lower, ate,
-                                        ate_upper, ate_lower), axis=1)
-        cols = ['upper', 'effects', 'lower', 'ate', 'ate_upper', 'ate_lower']
+                                        ate_upper, ate_lower, z_values_np),
+                                       axis=1)
+        cols = ['upper', 'effects', 'lower', 'ate', 'ate_upper', 'ate_lower',
+                'z_values']
     else:
-        cols = ['upper', 'effects', 'lower', 'ate']
-        effects_et_al = np.concatenate((upper, effects, lower, ate), axis=1)
+        cols = ['upper', 'effects', 'lower', 'ate', 'z_values']
+        effects_et_al = np.concatenate((upper, effects, lower, ate,
+                                        z_values_np), axis=1)
     datasave = pd.DataFrame(data=effects_et_al, columns=cols)
     gp.delete_file_if_exists(file_name_csv)
     datasave.to_csv(file_name_csv, index=False)
